use std::io::Cursor;

use gix::{bstr::BString, objs::Kind, traverse::tree::Recorder, Blob, Object, Repository};
use zip::{write::SimpleFileOptions, ZipWriter};

use crate::utils::error::{fail, Result};

pub struct ArchiveBuilder {
    writer: ZipWriter<Cursor<Vec<u8>>>,
}

impl std::default::Default for ArchiveBuilder {
    fn default() -> Self {
        Self {
            writer: ZipWriter::new(Cursor::new(Vec::new())),
        }
    }
}

impl std::fmt::Debug for ArchiveBuilder {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "<ArchiveBuilder ..>")
    }
}

impl ArchiveBuilder {
    pub fn new() -> Self {
        Self::default()
    }

    pub fn finish(mut self) -> Result<Vec<u8>> {
        Ok(self.writer.finish()?.into_inner())
    }

    pub fn archive(self, repo: &Repository, rev: &str) -> Result<Self> {
        let id = repo.rev_parse_single(rev)?;
        let obj = repo.find_object(id)?;
        self.archive_object(repo, obj)
    }

    pub fn archive_object(mut self, repo: &Repository, obj: Object<'_>) -> Result<Self> {
        let tree = match obj.kind {
            Kind::Blob => fail!("cannot archive single blobs"),
            Kind::Commit => {
                let tree = obj.try_into_commit()?.decode()?.tree();
                repo.find_object(tree)?.try_into_tree()?
            }
            Kind::Tag => {
                let tag = obj.try_into_tag()?.decode()?.target();
                repo.find_object(tag)?.try_into_tree()?
            }
            Kind::Tree => obj.try_into_tree()?,
        };
        let mut entries = Recorder::default();
        tree.traverse().breadthfirst(&mut entries)?;

        for entry in entries.records {
            if entry.mode.is_blob_or_symlink() {
                let permission = if entry.mode.is_executable() {
                    0o755
                } else {
                    0o644
                };
                self.archive_blob(
                    &entry.filepath,
                    repo.find_object(entry.oid)?.try_into_blob()?,
                    permission,
                )?;
            }
        }

        Ok(self)
    }

    fn archive_blob(&mut self, filepath: &BString, blob: Blob<'_>, permission: u32) -> Result<()> {
        let filepath = std::str::from_utf8(filepath)?;
        self.writer.start_file(
            filepath,
            SimpleFileOptions::default()
                .unix_permissions(permission)
                .compression_method(zip::CompressionMethod::Deflated),
        )?;
        std::io::copy(&mut Cursor::new(&blob.data), &mut self.writer)?;
        Ok(())
    }
}
