use gix::{
    objs::{tree::EntryMode, Kind},
    traverse::tree::{Recorder, Visit},
    Object, Repository, Tree,
};
use minijinja::value::Value;
use serde::Serialize;
use time::format_description;

use crate::{
    formats,
    utils::{
        build_path_rev, context_add,
        error::{error, fail, Result},
        paths::{get_parent_path, join_path, ObjectLocation},
        Context,
    },
    App,
};

const MAX_HIGHLIGHT_LEN: usize = 5 * 1024 * 1024;

impl App {
    pub fn add_navigation_context(
        &self,
        ctx: &mut Context,
        project: &str,
        rev: &str,
    ) -> Result<()> {
        context_add!(ctx,
            branches => self.get_branches(project)?,
            commits => Value::from_serialize(self.get_commits(project, rev)?),
            projects => &self.projects,
        );
        Ok(())
    }

    pub fn add_view_context(
        &self,
        ctx: &mut Context,
        loc: ObjectLocation<'_>,
        raw: bool,
    ) -> Result<()> {
        let repo = self
            .repos
            .get(loc.project)
            .ok_or_else(|| error!("No repo {project}", project = loc.project))?
            .to_thread_local();

        let path_rev = build_path_rev(loc.rev, loc.path);
        let obj = self.find_object(&repo, &path_rev)?;

        context_add!(
            ctx,
            current_rev => loc.rev,
            current_path => loc.path,
            current_project => loc.project,
        );

        match obj.kind {
            Kind::Blob => self.add_view_context_blob(ctx, &repo, loc, obj, raw),
            Kind::Commit => {
                self.add_view_context_tree(ctx, &repo, loc, obj.try_into_commit()?.tree()?)
            }
            Kind::Tree => self.add_view_context_tree(ctx, &repo, loc, obj.try_into_tree()?),
            kind => fail!("Cannot handle object of kind {kind}"),
        }
    }

    fn add_view_context_blob(
        &self,
        ctx: &mut Context,
        repo: &Repository,
        loc: ObjectLocation<'_>,
        obj: Object,
        raw: bool,
    ) -> Result<()> {
        context_add!(ctx, type => "blob");

        self.add_view_context_blob_content(ctx, loc, obj, raw)?;

        let parent_path = get_parent_path(loc.path).unwrap_or_default();
        let parent = self.find_tree(repo, &build_path_rev(loc.rev, parent_path))?;
        self.add_view_context_tree_children(ctx, parent_path, parent)?;

        Ok(())
    }

    fn add_view_context_tree(
        &self,
        ctx: &mut Context,
        repo: &Repository,
        loc: ObjectLocation<'_>,
        tree: Tree,
    ) -> Result<()> {
        let children = self.add_view_context_tree_children(ctx, loc.path, tree)?;

        if let Some(readme_path) = self.find_readme_path_for_tree(&children) {
            let readme_obj = self.find_blob(repo, &build_path_rev(loc.rev, readme_path))?;
            self.add_view_context_blob_content(ctx, loc.with_path(readme_path), readme_obj, false)?;
        }

        Ok(())
    }

    fn find_readme_path_for_tree<'children>(
        &self,
        children: &'children [TreePath],
    ) -> Option<&'children str> {
        for child in children {
            if matches!(child.name.as_str(), "Readme.md") {
                return Some(&child.path);
            }
        }
        None
    }

    fn add_view_context_tree_children(
        &self,
        ctx: &mut Context,
        path: &str,
        tree: Tree,
    ) -> Result<Vec<TreePath>> {
        let mut recorder = ShallowRecorder::new();
        tree.traverse().breadthfirst(&mut recorder)?;
        let recorder = recorder.into_inner();

        let mut children = Vec::new();
        for entry in recorder.records {
            // NOTE: entry.oid => lookup the entry
            let filepath = entry.filepath.to_string();
            children.push(TreePath {
                mode: entry_mode_to_string(entry.mode),
                path: join_path(&[path, &filepath]),
                name: filepath,
            });
        }
        children.sort_by_key(|obj| (obj.mode != "Tree", obj.path.to_lowercase()));

        context_add!(
            ctx,
            children => children,
            tree_path => path,
            parent_path => get_parent_path(path),
        );

        Ok(children)
    }

    fn add_view_context_blob_content(
        &self,
        ctx: &mut Context,
        loc: ObjectLocation<'_>,
        obj: Object,
        raw: bool,
    ) -> Result<()> {
        let blob_content = std::str::from_utf8(obj.data.as_slice())?;

        let content = if raw {
            blob_content.to_string()
        } else if let Some(res) = self.maybe_render_markdown(loc, blob_content)? {
            res
        } else if let Some(res) = self.maybe_render_notebook(loc, blob_content)? {
            res
        } else if let Some(res) = self.maybe_highlight(loc, blob_content)? {
            res
        } else {
            self.render_raw(blob_content)?
        };

        context_add!(ctx, content => content);
        Ok(())
    }
}

impl App {
    fn maybe_render_markdown(&self, loc: ObjectLocation<'_>, blob: &str) -> Result<Option<String>> {
        if !loc.path.ends_with(".md") {
            return Ok(None);
        }

        let base_url = loc.get_base_url(&self.prefix)?;
        let renderer = formats::markdown::HtmlRenderer::new(base_url).with_level_shift(1);
        Ok(renderer.render_str(blob).ok())
    }

    fn maybe_render_notebook(&self, loc: ObjectLocation<'_>, blob: &str) -> Result<Option<String>> {
        if !loc.path.ends_with(".ipynb") {
            return Ok(None);
        }

        let base_url = loc.get_base_url(&self.prefix)?;
        let renderer =
            formats::notebook::NotebookRenderer::new(base_url, &self.syntax_set, &self.theme_set);

        match renderer.render_str(blob) {
            Ok(result) => Ok(Some(result)),
            Err(err) => {
                println!(
                    "Error while rendering notebook {path}: {err}",
                    path = loc.path
                );
                Ok(None)
            }
        }
    }

    fn maybe_highlight(&self, loc: ObjectLocation<'_>, blob: &str) -> Result<Option<String>> {
        if blob.len() > MAX_HIGHLIGHT_LEN {
            return Ok(None);
        }

        let Some((_, ext)) = loc.path.rsplit_once('.') else {
            return Ok(None);
        };
        let Some(syntax_ref) = self.syntax_set.find_syntax_by_extension(ext) else {
            return Ok(None);
        };

        let Ok(result) = syntect::html::highlighted_html_for_string(
            blob,
            &self.syntax_set,
            syntax_ref,
            &self.theme_set.themes["InspiredGitHub"],
        ) else {
            return Ok(None);
        };

        let result = format!("<pre><code>{result}</code></pre>");
        Ok(Some(result))
    }

    fn render_raw(&self, blob_content: &str) -> Result<String> {
        let mut content = String::new();
        pulldown_cmark::escape::escape_html(&mut content, blob_content)?;
        Ok(format!("<pre><code>{content}</code></pre>"))
    }
}

impl App {
    fn get_branches(&self, project: &str) -> Result<Vec<String>> {
        let repo = self
            .repos
            .get(project)
            .ok_or_else(|| error!("No repo {project}"))?
            .to_thread_local();
        let refs = repo.references()?;
        let mut result = Vec::new();

        for branch in refs.local_branches()? {
            let branch = branch.map_err(|err| error!("{err}"))?;
            let branch = branch.name().as_bstr().to_string();
            let branch = branch
                .strip_prefix("refs/heads/")
                .ok_or_else(|| error!("Unexpected branch name"))?;

            result.push(branch.to_owned());
        }

        result.sort();
        Ok(result)
    }

    fn get_commits(&self, project: &str, rev: &str) -> Result<Vec<Commit>> {
        let repo = self
            .repos
            .get(project)
            .ok_or_else(|| error!("No repo {project}"))?
            .to_thread_local();
        let id = repo.rev_parse_single(rev)?;

        let format = format_description::parse("[year]-[month]-[day] [hour]:[minute]:[second]")?;

        let mut result = Vec::new();
        for parent in id.ancestors().all()? {
            let parent = parent?;
            let commit = parent.object()?;
            // let commit = object.try_into_commit()?;
            let message = commit.message()?;
            let time = commit.time()?;

            let time = time::OffsetDateTime::from_unix_timestamp(time.seconds)?;
            let time = time.format(&format)?;

            result.push(Commit {
                id: parent.id.to_hex().to_string(),
                title: message.title.to_string(),
                time,
            });
        }

        Ok(result)
    }

    fn find_object<'repo>(&self, repo: &'repo Repository, rev: &str) -> Result<Object<'repo>> {
        let id = repo.rev_parse_single(rev)?;
        let obj = repo.find_object(id)?;
        Ok(obj)
    }

    fn find_tree<'repo>(&self, repo: &'repo Repository, rev: &str) -> Result<Tree<'repo>> {
        let obj = self.find_object(repo, rev)?;
        match obj.kind {
            Kind::Commit => Ok(obj.try_into_commit()?.tree()?),
            Kind::Tree => Ok(obj.try_into_tree()?),
            kind => fail!("Cannot convert object of kind {kind} to tree"),
        }
    }

    fn find_blob<'repo>(&self, repo: &'repo Repository, rev: &str) -> Result<Object<'repo>> {
        let obj = self.find_object(repo, rev)?;
        match obj.kind {
            Kind::Blob => Ok(obj),
            kind => fail!("Cannot convert object of kind {kind} to blob"),
        }
    }
}

#[derive(Debug, Serialize)]
struct Commit {
    id: String,
    title: String,
    time: String,
}

#[derive(Debug, Default)]
struct ShallowRecorder {
    recorder: Recorder,
}

impl ShallowRecorder {
    pub fn new() -> Self {
        Self::default()
    }

    pub fn into_inner(self) -> Recorder {
        self.recorder
    }
}

impl Visit for ShallowRecorder {
    fn pop_front_tracked_path_and_set_current(&mut self) {
        self.recorder.pop_front_tracked_path_and_set_current()
    }

    fn pop_path_component(&mut self) {
        self.recorder.pop_path_component()
    }

    fn push_back_tracked_path_component(&mut self, component: &gix::bstr::BStr) {
        self.recorder.push_back_tracked_path_component(component)
    }

    fn push_path_component(&mut self, component: &gix::bstr::BStr) {
        self.recorder.push_path_component(component)
    }

    fn visit_nontree(
        &mut self,
        entry: &gix::objs::tree::EntryRef<'_>,
    ) -> gix::traverse::tree::visit::Action {
        self.recorder.visit_nontree(entry)
    }

    fn visit_tree(
        &mut self,
        entry: &gix::objs::tree::EntryRef<'_>,
    ) -> gix::traverse::tree::visit::Action {
        self.recorder.visit_tree(entry);
        gix::traverse::tree::visit::Action::Skip
    }
}

#[derive(Debug, Serialize)]
struct TreePath {
    mode: &'static str,
    path: String,
    name: String,
}

fn entry_mode_to_string(mode: EntryMode) -> &'static str {
    if mode.is_blob() {
        "Blob"
    } else if mode.is_link() {
        "Link"
    } else if mode.is_tree() {
        "Tree"
    } else if mode.is_commit() {
        "Commit"
    } else {
        "Unknown"
    }
}
