pub mod archive_builder;
mod context;
mod router;

use std::{collections::HashMap, path::PathBuf};

use gix::ThreadSafeRepository;
use minijinja::Environment;
use serde::Deserialize;
use syntect::{highlighting::ThemeSet, parsing::SyntaxSet};

use crate::utils::error::{error, fail, Result};

#[derive(Debug)]
pub struct Config {
    pub repos: HashMap<String, PathBuf>,
    pub projects: Vec<String>,
    pub default_project: String,
    pub port: u16,
    pub default_branch: String,
    pub prefix: String,
}

impl Config {
    pub fn from_env() -> Result<Config> {
        let mut config = envy::prefixed("GIT_BROWSER_").from_env::<RawConfig>()?;

        for arg in std::env::args().skip(1) {
            let new_cfg = std::fs::read_to_string(arg)?;
            let new_cfg: RawConfig = toml::from_str(&new_cfg)?;

            config = config.merge(new_cfg);
        }

        let config = config.normalize()?;
        Ok(config)
    }
}

#[derive(Debug, Deserialize)]
struct RawConfig {
    repos: Option<Vec<String>>,
    default_repo: Option<String>,
    port: Option<u16>,
    default_branch: Option<String>,
    prefix: Option<String>,
}

impl RawConfig {
    pub fn merge(self, other: RawConfig) -> Self {
        Self {
            repos: other.repos.or(self.repos),
            default_repo: other.default_repo.or(self.default_repo),
            port: other.port.or(self.port),
            default_branch: other.default_branch.or(self.default_branch),
            prefix: other.prefix.or(self.prefix),
        }
    }

    pub fn normalize(self) -> Result<Config> {
        let mut repos = HashMap::new();
        let mut projects = Vec::new();

        for repo in self.repos.unwrap_or_else(|| vec![String::from(".")]) {
            let repo = PathBuf::from(&repo).canonicalize()?;
            let name = repo
                .file_name()
                .ok_or_else(|| error!("Canonicalized path without filename"))?
                .to_str()
                .ok_or_else(|| error!("Path not representable as UTF8"))?
                .to_owned();

            projects.push(name.clone());
            repos.insert(name, repo);
        }

        if repos.is_empty() {
            fail!("No repositories specified");
        }

        let default_project = self.default_repo.unwrap_or_else(|| projects[0].clone());
        let port = self.port.unwrap_or(9090);
        let default_branch = self.default_branch.unwrap_or_else(|| String::from("main"));
        let prefix = self.prefix.unwrap_or_default();

        if !prefix.is_empty() && (!prefix.starts_with('/') || prefix.ends_with('/')) {
            fail!("Prefix should start with '/' and not have a trailing '/'");
        }

        let config = Config {
            repos,
            projects,
            default_project,
            port,
            default_branch,
            prefix,
        };
        Ok(config)
    }
}

pub struct App {
    pub env: Environment<'static>,
    pub repos: HashMap<String, ThreadSafeRepository>,
    pub projects: Vec<String>,
    pub default_branch: String,
    pub default_project: String,
    pub prefix: String,
    pub addr: String,
    pub theme_set: ThemeSet,
    pub syntax_set: SyntaxSet,
}

impl App {
    pub fn new(config: Config) -> Result<Self> {
        let repos: Result<HashMap<String, ThreadSafeRepository>> = config
            .repos
            .into_iter()
            .map(|(name, path)| -> Result<(String, ThreadSafeRepository)> {
                Ok((name, ThreadSafeRepository::open(path)?))
            })
            .collect();
        let repos = repos?;

        let env = build_minijinja_env()?;

        let syntax_set = SyntaxSet::load_defaults_newlines();
        let theme_set = ThemeSet::load_defaults();

        for key in theme_set.themes.keys() {
            println!("{key}")
        }

        let app = App {
            env,
            repos,
            // TODO: allow to customize this from env
            default_branch: String::from("main"),
            default_project: config.default_project,
            projects: config.projects,
            prefix: config.prefix,
            addr: format!("0.0.0.0:{}", config.port),
            theme_set,
            syntax_set,
        };
        Ok(app)
    }
}

fn build_minijinja_env() -> Result<Environment<'static>> {
    use template_fragments as tf;

    let mut env = minijinja::Environment::new();

    for (path, template) in [("index.html", include_str!("../templates/index.html"))] {
        for (fragment_name, template_fragment) in tf::split_templates(template)? {
            let res = env.add_template_owned(
                tf::join_path(path, &fragment_name),
                template_fragment.clone(),
            );

            if let Err(err) = res {
                let template_fragment = add_line_numbers(&template_fragment);
                println!("Invalid template:\n###\n{template_fragment}\n###");
                return Err(err.into());
            }
        }
    }

    env.add_filter("str_slice", template::str_slice);
    env.add_filter("optional_path", template::optional_path);
    env.add_test("object_id", template::is_object_id);
    env.add_function("breadcrumbs", template::breadcrumbs);

    Ok(env)
}

fn add_line_numbers(s: &str) -> String {
    use std::fmt::Write;

    let mut res = String::new();
    for (idx, line) in s.lines().enumerate() {
        writeln!(&mut res, "{idx:3} {line}").unwrap();
    }
    res
}

mod template {
    use minijinja::value::Value;
    use serde::Serialize;

    pub fn str_slice(value: String, start: usize, end: usize) -> String {
        value[start..end].to_string()
    }

    pub fn optional_path(path: String) -> String {
        if path.is_empty() {
            String::new()
        } else if !path.starts_with('/') {
            format!("/:/{path}")
        } else {
            format!("/:{path}")
        }
    }

    pub fn is_object_id(s: String) -> bool {
        (s.len() == 40) && s.chars().all(|x| x.is_alphanumeric())
    }

    pub fn breadcrumbs(path: String) -> Vec<Breadcrumb> {
        let mut res = Vec::new();

        let mut prev = 0;
        for (pos, sep) in path.match_indices('/') {
            if !path[prev..pos].is_empty() {
                res.push(Breadcrumb {
                    path: path[..pos].to_string(),
                    fragment: path[prev..pos].to_string(),
                });
            }
            // skip the separator itself
            prev = pos + sep.len();
        }

        if !path[prev..].is_empty() {
            res.push(Breadcrumb {
                path: path.to_string(),
                fragment: path[prev..].to_string(),
            });
        }

        res
    }

    #[derive(Debug, Serialize)]
    pub struct Breadcrumb {
        path: String,
        fragment: String,
    }

    impl From<Breadcrumb> for Value {
        fn from(value: Breadcrumb) -> Self {
            Value::from_serialize(value)
        }
    }
}
