use std::{collections::HashMap, sync::Arc};

use axum::{
    extract::{Path, Query, State},
    http::{HeaderMap, HeaderValue, Response, StatusCode},
    response::IntoResponse,
    routing::get,
    Router,
};
use minijinja::Value;

use crate::{
    app::{archive_builder::ArchiveBuilder, App},
    utils::{
        context_add,
        error::{error, fail, Error, Result},
        paths::{join_path, ObjectLocation},
        Context,
    },
};

impl App {
    pub fn into_router(self: Arc<App>) -> Router {
        let router = Router::new();

        let router = if !self.prefix.is_empty() {
            router.route(&self.prefix, get(handle_index))
        } else {
            router
        };

        router
            .route(&format!("{}/", self.prefix), get(handle_index))
            .route(
                &format!("{}/:project", self.prefix),
                get(handle_view_project),
            )
            .route(
                &format!("{}/:project/", self.prefix),
                get(handle_view_project),
            )
            .route(
                &format!("{}/:project/*path_with_rev", self.prefix),
                get(handle_view),
            )
            .route(
                &format!("{}/:project/download/*rev", self.prefix),
                get(handle_download),
            )
            .route(&format!("{}/static/:asset", self.prefix), get(serve_static))
            .with_state(self)
    }
}

async fn serve_static(Path(path): Path<String>) -> impl IntoResponse {
    fn response(data: &str, mime_type: &'static str) -> Response<String> {
        let mut response = Response::new(data.to_string());
        let headers = response.headers_mut();
        headers.insert("Content-Type", HeaderValue::from_static(mime_type));

        response
    }

    match path.as_str() {
        "htmx.min.js" => response(
            include_str!("../static/htmx.min.js"),
            "application/javascript",
        ),
        "htmx.ext.path-params.js" => response(
            include_str!("../static/htmx.ext.path-params.js"),
            "application/javascript",
        ),
        "app.css" => response(include_str!("../static/app.css"), "text/css"),
        _ => {
            let mut response = Response::new("unknown file".to_string());
            *response.status_mut() = StatusCode::NOT_FOUND;
            response
        }
    }
}

async fn handle_index(
    State(app): State<Arc<App>>,
    headers: HeaderMap,
    query: Query<HashMap<String, String>>,
) -> impl IntoResponse {
    handle_view(
        Path((
            app.default_project.to_owned(),
            app.default_branch.to_owned(),
        )),
        State(app),
        headers,
        query,
    )
    .await
}

async fn handle_view_project(
    Path(project_name): Path<String>,
    State(app): State<Arc<App>>,
    headers: HeaderMap,
    query: Query<HashMap<String, String>>,
) -> impl IntoResponse {
    handle_view(
        Path((project_name, app.default_branch.to_owned())),
        State(app),
        headers,
        query,
    )
    .await
}

async fn handle_view(
    Path((project, path_with_rev)): Path<(String, String)>,
    State(app): State<Arc<App>>,
    headers: HeaderMap,
    query: Query<HashMap<String, String>>,
) -> Result<impl IntoResponse> {
    let (rev, path) = split_path_with_rev(&path_with_rev);
    let htmx_target = headers.get("HX-Target");
    let htmx_target = htmx_target.map(|value| value.to_str()).transpose()?;
    let raw = query.contains_key("raw");

    let mut ctx = Context::new();
    if matches!(htmx_target, None | Some("content")) {
        app.add_navigation_context(&mut ctx, &project, &rev)?;
    }

    app.add_view_context(&mut ctx, ObjectLocation::new(&project, &rev, &path), raw)?;
    context_add!(&mut ctx, prefix => app.prefix);

    // TODO: skip unnecessary work in context prep?
    if raw {
        return Ok(Response::new(app.env.render_str("{{ content }}", ctx)?));
    }

    let content = match htmx_target {
        Some("main") => app.render_template("index.html#main", ctx)?,
        Some("content") => app.render_template("index.html#content", ctx)?,
        Some(target) => fail!("Unknown htmx target: {target}"),
        None => app.render_template("index.html", ctx)?,
    };
    Ok(Response::new(content))
}

async fn handle_download(
    Path((project, rev)): Path<(String, String)>,
    State(app): State<Arc<App>>,
) -> Result<impl IntoResponse> {
    let repo = app
        .repos
        .get(&project)
        .ok_or_else(|| error!("Could not find repository {project}"))?
        .to_thread_local();
    let data = ArchiveBuilder::new().archive(&repo, &rev)?.finish()?;

    Ok(([("Content-Type", "application/zip")], data))
}

impl IntoResponse for Error {
    fn into_response(self) -> axum::response::Response {
        StatusCode::INTERNAL_SERVER_ERROR.into_response()
    }
}

fn split_path_with_rev(path_with_rev: &str) -> (String, String) {
    let path_with_rev = path_with_rev.strip_prefix(':').unwrap_or(path_with_rev);

    let path_with_rev = path_with_rev.split('/').collect::<Vec<_>>();
    let path_with_rev = path_with_rev.as_slice();

    let (rev, path);
    if let Some(pos) = path_with_rev.iter().position(|&p| p == ":") {
        rev = &path_with_rev[..pos];
        path = &path_with_rev[pos + 1..];
    } else {
        rev = path_with_rev;
        path = &[];
    }
    (join_path(rev), join_path(path))
}

impl App {
    pub fn render_template(&self, template: &str, ctx: Context) -> Result<String> {
        let template = self.env.get_template(template)?;
        let content = template.render(Value::from(ctx))?;
        Ok(content)
    }
}
