use std::fmt::Write;

use pulldown_cmark::{escape, Event, Options, Parser, Tag};

use crate::utils::{
    error::{fail, Result},
    paths::normalize_path,
};

#[derive(Default)]
pub struct HtmlRenderer {
    pub level_shift: i32,
    pub base_url: String,
}

impl HtmlRenderer {
    pub fn new(base_url: String) -> Self {
        Self {
            level_shift: 0,
            base_url,
        }
    }

    pub fn with_level_shift(mut self, value: i32) -> Self {
        self.level_shift = value;
        self
    }

    pub fn modify<F: FnOnce(&mut HtmlRenderer)>(mut self, f: F) -> Self {
        f(&mut self);
        self
    }

    pub fn render_str(&self, source: &str) -> Result<String> {
        let mut res = String::new();
        self.render(
            &mut res,
            Parser::new_ext(source, Options::ENABLE_TASKLISTS | Options::ENABLE_TABLES),
        )?;
        Ok(res)
    }

    pub fn render<'event, Iter>(&self, target: &mut String, iter: Iter) -> Result<()>
    where
        Iter: IntoIterator<Item = Event<'event>>,
    {
        render(target, iter, self)?;
        Ok(())
    }
}

fn render<'event, Iter>(target: &mut String, iter: Iter, options: &HtmlRenderer) -> Result<()>
where
    Iter: IntoIterator<Item = Event<'event>>,
{
    use Event::*;
    use Tag::*;

    for item in iter {
        match item {
            Start(Heading(level, _, _)) => {
                let level = i32::clamp(level as i32 + options.level_shift, 1, 6);
                write!(target, "<h{level}>")?
            }
            End(Heading(level, _, _)) => {
                let level = i32::clamp(level as i32 + options.level_shift, 1, 6);
                writeln!(target, "</h{level}>")?
            }

            Start(Paragraph) => write!(target, "<p>")?,
            End(Paragraph) => writeln!(target, "</p>")?,
            Start(Link(_, href, _)) => {
                let is_anchored = href.contains("://");

                let href = escape_href(&href)?;
                if is_anchored {
                    write!(target, "<a href=\"{href}\" target=\"_blank\">")?
                } else {
                    let href = normalize_path(&href)?;
                    write!(
                        target,
                        "<a href=\"{base_url}/{href}\">",
                        base_url = options.base_url
                    )?
                }
            }
            End(Link(_, _, _)) => write!(target, "</a>")?,
            // TODO: support numbered list
            Start(List(_)) => writeln!(target, "<ul>")?,
            End(List(_)) => writeln!(target, "</ul>")?,
            Start(Item) => write!(target, "<li>")?,
            End(Item) => writeln!(target, "</li>")?,
            Start(CodeBlock(_)) => write!(target, "<pre><code>")?,
            End(CodeBlock(_)) => write!(target, "</code></pre>")?,
            Start(BlockQuote) => write!(target, "<blockquote>")?,
            End(BlockQuote) => writeln!(target, "</blockquote>")?,

            Text(text) => {
                let text = escape_html(&text)?;
                write!(target, "{text}")?
            }
            Html(html) => write!(target, "{html}")?,
            SoftBreak => writeln!(target)?,
            HardBreak => writeln!(target, "<br>")?,

            Code(code) => {
                let code = escape_html(&code)?;
                write!(target, "<code>{code}</code>")?
            }
            Rule => writeln!(target, "<hr/>")?,

            // NOTE: the space is required for correct rendering
            TaskListMarker(true) => write!(target, "<input type=\"checkbox\" disabled checked/> ")?,
            TaskListMarker(false) => write!(target, "<input type=\"checkbox\" disabled/> ")?,

            Start(Image(_, href, _)) => {
                let href = escape_href(&href)?;
                write!(target, "<img src=\"{href}\"/>")?
            }
            End(Image(_, _, _)) => {}

            // TODO: handle table alignment + correct header
            Start(Table(_)) => writeln!(target, "<table>")?,
            End(Table(_)) => writeln!(target, "</table>")?,
            Start(TableHead) => writeln!(target, "<tr>")?,
            End(TableHead) => writeln!(target, "</tr>")?,
            Start(TableRow) => writeln!(target, "<tr>")?,
            End(TableRow) => writeln!(target, "</tr>")?,
            Start(TableCell) => write!(target, "<td>")?,
            End(TableCell) => write!(target, "</td>")?,

            Start(Emphasis) => write!(target, "<em>")?,
            End(Emphasis) => write!(target, "</em>")?,
            Start(Strong) => write!(target, "<strong>")?,
            End(Strong) => write!(target, "</strong>")?,
            Start(Strikethrough) => write!(target, "<del>")?,
            End(Strikethrough) => write!(target, "<del>")?,

            Start(FootnoteDefinition(_)) | End(FootnoteDefinition(_)) | FootnoteReference(_) => {
                fail!("Footnotes are not supported")
            }
        }
    }

    Ok(())
}

pub fn escape_html(s: &str) -> Result<String> {
    let mut res = String::new();
    escape::escape_html(&mut res, s)?;
    Ok(res)
}

pub fn escape_href(s: &str) -> Result<String> {
    let mut res = String::new();
    escape::escape_href(&mut res, s)?;
    Ok(res)
}
