use std::{collections::HashMap, fmt::Write};

use serde::Deserialize;
use serde_json::Value;
use syntect::{highlighting::ThemeSet, parsing::SyntaxSet};

use crate::utils::error::{error, Result};

use super::markdown;

pub struct NotebookRenderer<'a> {
    md_renderer: markdown::HtmlRenderer,
    syntax_set: &'a SyntaxSet,
    theme_set: &'a ThemeSet,
}

impl<'a> NotebookRenderer<'a> {
    pub fn new(base_url: String, syntax_set: &'a SyntaxSet, theme_set: &'a ThemeSet) -> Self {
        Self {
            md_renderer: markdown::HtmlRenderer::new(base_url).with_level_shift(1),
            syntax_set,
            theme_set,
        }
    }

    pub fn render_str(&self, src: &str) -> Result<String> {
        let notebook: Notebook = serde_json::from_str(src)?;

        let mut res = String::new();
        for cell in &notebook.cells {
            render_cell(
                &mut res,
                cell,
                &self.md_renderer,
                self.syntax_set,
                self.theme_set,
            )?;
        }
        Ok(res)
    }
}

pub fn render_cell(
    res: &mut String,
    cell: &Cell,
    md_renderer: &markdown::HtmlRenderer,
    syntax_set: &SyntaxSet,
    theme_set: &ThemeSet,
) -> Result<()> {
    match cell {
        Cell::Markdown(cell) => {
            let source = join(&cell.source, "\n");
            let rendered = md_renderer.render_str(&source)?;

            write!(res, "<div class=\"cell markdown\">")?;
            write!(res, "{rendered}")?;
            write!(res, "</div>")?;
        }
        Cell::Code(cell) => {
            let source = join(&cell.source, "\n");
            let syntax_ref = syntax_set
                .find_syntax_by_extension("py")
                .ok_or_else(|| error!("Cannot find Python syntax"))?;
            let rendered = syntect::html::highlighted_html_for_string(
                &source,
                syntax_set,
                syntax_ref,
                &theme_set.themes["InspiredGitHub"],
            )?;

            write!(res, "<div class=\"cell code\">")?;
            writeln!(res, "<div class=\"source\">")?;
            writeln!(res, "<pre><code>{rendered}</code></pre>")?;
            writeln!(res, "</div>")?;
            writeln!(res, "<div class=\"outputs\">")?;
            for output in &cell.outputs {
                render_output(res, output)?;
            }
            writeln!(res, "</div>")?;
            write!(res, "</div>")?;
        }
        Cell::Unknown => {
            write!(res, "<div class=\"cell unknown\">")?;
            write!(res, "<div>unknown cell</div>")?;
            write!(res, "</div>")?;
        }
    }
    Ok(())
}

pub fn render_output(res: &mut String, output: &Output) -> Result<()> {
    match output {
        Output::Stream(output) => {
            let data = join(&output.text, "\n");
            writeln!(res, "<pre><code>{data}</code></pre>")?;
        }
        Output::ExecuteResult(ExecuteResultOutput { data })
        | Output::DisplayData(DisplayDataOutput { data }) => {
            if let Some(image) = data.get("image/png") {
                let data = join(&image.0, "");
                writeln!(res, "<img src=\"data:image/png;base64,{}\"/>", data)?;
            } else if let Some(data) = data.get("text/plain") {
                let source = join(&data.0, "\n");
                writeln!(res, "<pre><code>{source}</code></pre>")?;
            }
        }
        Output::Unknown => {
            write!(res, "<div>unknown output</div>")?;
        }
    }

    Ok(())
}

fn join(lines: &[String], sep: &str) -> String {
    let mut res = String::new();
    for line in lines {
        res.push_str(line);
        res.push_str(sep);
    }
    res
}

#[derive(Debug, Deserialize)]
pub struct Notebook {
    pub cells: Vec<Cell>,
}

#[derive(Debug, Deserialize)]
#[serde(tag = "cell_type")]
pub enum Cell {
    #[serde(rename = "markdown")]
    Markdown(Markdown),
    #[serde(rename = "code")]
    Code(Code),
    #[serde(other)]
    Unknown,
}

#[derive(Debug, Deserialize)]
pub struct Markdown {
    pub source: Vec<String>,
}

#[derive(Debug, Deserialize)]
pub struct Code {
    pub source: Vec<String>,
    pub outputs: Vec<Output>,
}

#[derive(Debug, Deserialize)]
#[serde(tag = "output_type")]
pub enum Output {
    #[serde(rename = "stream")]
    Stream(StreamOutput),
    #[serde(rename = "display_data")]
    DisplayData(DisplayDataOutput),
    #[serde(rename = "execute_result")]
    ExecuteResult(ExecuteResultOutput),
    #[serde(other)]
    Unknown,
}

#[derive(Debug, Deserialize)]
pub struct StreamOutput {
    text: Vec<String>,
}

#[derive(Debug, Deserialize)]
pub struct DisplayDataOutput {
    data: HashMap<String, MultiLineString>,
}

#[derive(Debug, Deserialize)]
pub struct ExecuteResultOutput {
    // TODO: fix deserialization
    data: HashMap<String, MultiLineString>,
}

#[derive(Debug, Deserialize)]
#[serde(from = "MultiLineStringImpl")]
pub struct MultiLineString(Vec<String>);

#[derive(Debug, Deserialize)]
#[serde(untagged)]
enum MultiLineStringImpl {
    String(String),
    Lines(Vec<String>),
    #[allow(dead_code)]
    Value(Value),
}

impl From<MultiLineStringImpl> for MultiLineString {
    fn from(value: MultiLineStringImpl) -> Self {
        match value {
            MultiLineStringImpl::String(val) => MultiLineString(vec![val]),
            MultiLineStringImpl::Lines(lines) => MultiLineString(lines),
            MultiLineStringImpl::Value(_) => MultiLineString(vec![]),
        }
    }
}
