use super::error::{error, Result};

pub fn join_path(parts: &[&str]) -> String {
    let mut path = String::new();
    for &p in parts {
        if p.is_empty() {
            continue;
        }
        if !path.is_empty() {
            path.push('/');
        }
        path.push_str(p);
    }
    path
}

pub fn get_parent_path(path: &str) -> Option<&str> {
    if path.is_empty() {
        None
    } else if let Some((parent_path, _)) = path.rsplit_once('/') {
        Some(parent_path)
    } else {
        Some("")
    }
}

/// Normalize a path by removing `"."` and `".."`
///
/// The resulting path will always have the leading and trailing `"/"` stripped.
/// Empty path fragments, e.g., `"//"`, will be stripped as well.
///
/// This function fails if the resulting path would point outside the root
pub fn normalize_path(path: &str) -> Result<String> {
    let mut stack = Vec::new();

    for p in path.split('/') {
        if p == "." {
            continue;
        } else if p == ".." {
            stack
                .pop()
                .ok_or_else(|| error!("invalid path: traverses outside root"))?;
        } else if !p.is_empty() {
            stack.push(p.to_owned());
        }
    }

    let mut res = String::new();
    for (idx, p) in stack.iter().enumerate() {
        if idx == 0 {
            res.push_str(p);
        } else {
            res.push('/');
            res.push_str(p);
        }
    }
    Ok(res)
}

/// Represent the full location of an object
#[derive(Debug, Clone, Copy)]
pub struct ObjectLocation<'a> {
    pub project: &'a str,
    pub rev: &'a str,
    pub path: &'a str,
}

impl<'a> ObjectLocation<'a> {
    pub fn new(project: &'a str, rev: &'a str, path: &'a str) -> Self {
        Self { project, rev, path }
    }

    /// Change the path
    pub fn with_path<'b>(self, path: &'b str) -> ObjectLocation<'b>
    where
        'a: 'b,
    {
        ObjectLocation::<'b> {
            project: self.project,
            rev: self.rev,
            path,
        }
    }
}

impl<'a> ObjectLocation<'a> {
    /// Get the url for renderable documents that is used as the anchor for
    /// relative links.
    pub fn get_base_url(&self, prefix: &str) -> Result<String> {
        let parent_path = get_parent_path(self.path)
            .ok_or_else(|| error!("Cannot get base_url for empty path"))?;
        if !parent_path.is_empty() {
            Ok(format!(
                "{prefix}/{project}/:/{rev}/:/{parent_path}",
                project = self.project,
                rev = self.rev
            ))
        } else {
            Ok(format!(
                "{prefix}/{project}/:/{rev}/:",
                project = self.project,
                rev = self.rev
            ))
        }
    }
}

#[cfg(test)]
mod test {
    mod normalize_path {
        use super::super::normalize_path;
        use crate::utils::testing::assert_matches;

        macro_rules! test {
            (
                name = $name:ident,
                path = $path:expr,
                expected = $expected:pat
            ) => {
                test!(name = $name, path = $path, expected = $expected,);
            };
            (
                name = $name:ident,
                path = $path:expr,
                expected = $expected:pat,
            ) => {
                #[test]
                fn $name() {
                    let actual = normalize_path($path);
                    assert_matches!(actual, $expected);
                }
            };
        }

        test!(name = invalid_path_1, path = "..", expected = Err(_));
        test!(name = invalid_path_2, path = "../foo", expected = Err(_));
        test!(name = invalid_path_3, path = "foo/../..", expected = Err(_));

        test!(name = empty_path, path = "", expected = Ok(""));
        test!(name = simple_path, path = "foo", expected = Ok("foo"));
        test!(
            name = simple_path_rooted,
            path = "/foo",
            expected = Ok("foo"),
        );
        test!(name = traversal, path = "foo/../bar", expected = Ok("bar"));
        test!(name = leading_dot, path = "./bar", expected = Ok("bar"));
    }

    mod object_location {
        mod get_base_url {
            use super::super::super::ObjectLocation;
            use crate::utils::testing::assert_matches;

            macro_rules! test {
                (
                    name = $name:ident,
                    prefix = $prefix:expr,
                    project = $project:expr,
                    rev = $rev:expr,
                    path = $path:expr,
                    expected = $expected:pat,
                ) => {
                    #[test]
                    fn $name() {
                        let actual =
                            ObjectLocation::new($project, $rev, $path).get_base_url($prefix);
                        assert_matches!(actual, $expected);
                    }
                };
            }

            test!(
                name = example,
                prefix = "http://localhost:9090",
                project = "git-browser",
                rev = "main",
                path = "Readme.md",
                expected = Ok("http://localhost:9090/git-browser/:/main/:"),
            );

            test!(
                name = example_with_subdir,
                prefix = "http://localhost:9090",
                project = "git-browser",
                rev = "main",
                path = "sub-dir/Readme.md",
                expected = Ok("http://localhost:9090/git-browser/:/main/:/sub-dir"),
            );
            test!(
                name = missing_path,
                prefix = "http://localhost:9090",
                project = "git-browser",
                rev = "main",
                path = "",
                expected = Err(_),
            );
        }
    }
}
